;;;; command.scm -- vt100 commands

;;; Created:    <2002-06-19 22:27:12 foof>
;;; Time-stamp: <2002-06-20 00:24:32 foof>
;;; Author:     Alex Shinn <foof@synthcode.com>


(define-module term.command
  (export color colored make-color-proc term-code)
  )
(select-module term.command)

;;; vt100 escape codes

(define escape-codes #f)

(cond ((not escape-codes)
       (set! escape-codes (make-hash-table 'eq?))
       (for-each
        (lambda (c)
          (let ((data (cddr c)))
            (hash-table-put! escape-codes (car c) data)
            (hash-table-put! escape-codes (cadr c) data)))
        '(

;; Mnemonic   Name               Sequence                        Default   Type
(APC  applicatn-program-command (#\esc Fe)                        #f        Delim)
(CBT  cursor-backward-tab       (#\esc #\[ Pn #\Z)                1         eF)
(CCH  cancel-previous-character (#\esc #\T)                       #f)
(CHA  cursor-horzntal-absolute  (#\esc #\[ Pn #\G)                1         eF)
(CHT  cursor-horizontal-tab     (#\esc #\[ Pn #\I)                1         eF)
(CNL  cursor-next-line          (#\esc #\[ Pn #\E)                1         eF)
(CPL  cursor-preceding-line     (#\esc #\[ Pn #\F)                1         eF)
(CPR  cursor-position-report    (#\esc #\[ Pn 2 #\R)              (1 1))
(CSI  control-sequence-intro    (#\esc #\[)                       #f        Intro)
(CTC  cursor-tab-control        (#\esc #\[ Ps #\W)                0         eF)

(CUB  cursor-backward           (#\esc #\[ Pn #\D)                1         eF)
(CUD  cursor-down               (#\esc #\[ Pn #\B)                1         eF)
(CUF  cursor-forward            (#\esc #\[ Pn #\C)                1         eF)
(CUP  cursor-position           (#\esc #\[ Pn 2 #\H)              (1 1)     eF)
(CUU  cursor-up                 (#\esc #\[ Pn #\A)                1         eF)
(CVT  cursor-vertical-tab       (#\esc #\[ Pn #\Y)                #f        eF)
(DA   device-attributes         (#\esc #\[ Pn #\c)                0)
(DAQ  define-area-qualification (#\esc #\[ Ps #\o)                0)
(DCH  delete-character          (#\esc #\[ Pn #\P)                1         eF)
(DCS  device-control-string     (#\esc #\P)                       #f        Delim)

(DL   delete-line               (#\esc #\[ Pn #\M)                1         eF)
(DMI  disable-manual-input      (#\esc #\\)                       #f        Fs)
(DSR  device-status-report      (#\esc #\[ Ps #\n)                0)
(EA   erase-in-area             (#\esc #\[ Ps #\O)                0         eF)
(ECH  erase-character           (#\esc #\[ Pn #\X)                1         eF)
(ED   erase-in-display          (#\esc #\[ Ps #\J)                0         eF)
(EF   erase-in-field            (#\esc #\[ Ps #\N)                0         eF)
(EL   erase-in-line             (#\esc #\[ Ps #\K)                0         eF)
(EMI  enable-manual-input       (#\esc #\b)                       #f        Fs)
(EPA  end-of-protected-area     (#\esc #\W)                       #f)

(ESA  end-of-selected-area      (#\esc #\G)                       #f)
(FNT  font-selection            (#\esc #\[ Pn 2 #\space #\D)      (0 0)     FE)
(GSM  graphic-size-modify       (#\esc #\[ Pn 2 #\space #\B)      (100 100) FE)
(GSS  graphic-size-selection    (#\esc #\[ Pn #\space #\C)        none      FE)
(HPA  horz-position-absolute    (#\esc #\[ Pn #\`)                1         FE)
(HPR  horz-position-relative    (#\esc #\[ Pn #\a)                1         FE)
(HTJ  horz-tab-w/justification  (#\esc #\I)                       #f        FE)
(HTS  horizontal-tab-set        (#\esc #\H)                       #f        FE)
(HVP  horz-&-vertical-position  (#\esc #\[ Pn 2 #\f)              (1 1)     FE)
(ICH  insert-character          (#\esc #\[ Pn #\@)                1         eF)

(IL   insert-line               (#\esc #\[ Pn #\L)                1         eF)
(IND  index                     (#\esc #\D)                       #f        FE)
(INT  interrupt                 (#\esc #\a)                       #f        Fs)
(JFY  justify                   (#\esc #\[ Ps * #\space #\F)      0         FE)
(MC   media-copy                (#\esc #\[ Ps #\i)                0)
(MW   message-waiting           (#\esc #\U)                       #f)
(NEL  next-line                 (#\esc #\E)                       #f        FE)
(NP   next-page                 (#\esc #\[ Pn #\U)                1         eF)
(OSC  operating-system-command  (#\esc #\])                       #f        Delim)
(PLD  partial-line-down         (#\esc #\K)                       #f        FE)

(PLU  partial-line-up           (#\esc #\L)                       #f        FE)
(PM   privacy-message           (#\esc #\^)                       #f        Delim)
(PP   preceding-page            (#\esc #\[ Pn #\V)                1         eF)
(PU1  private-use-1             (#\esc #\Q)                       #f)
(PU2  private-use-2             (#\esc #\R)                       #f)
(QUAD typographic-quadding      (#\esc #\[ Ps #\space #\H)        0         FE)
(REP  repeat-char-or-control    (#\esc #\[ Pn #\b)                1)
(RI   reverse-index             (#\esc #\M)                       #f        FE)
(RIS  reset-to-initial-state    (#\esc #\c)                       #f        Fs)
(RM   reset-mode                (#\esc #\[ Ps #\l)                none)

(SD   scroll-down               (#\esc #\[ Pn #\T)                1         eF)
(SEM  select-edit-extent-mode   (#\esc #\[ Ps #\Q)                0)
(SGR  select-graphic-rendition  (#\esc #\[ Ps #\m)                0         FE)
(SL   scroll-left               (#\esc #\[ Pn #\space #\@)        1         eF)
(SM   select-mode               (#\esc #\[ Ps #\h)                none)
(SPA  start-of-protected-area   (#\esc #\V)                       #f)
(SPI  spacing-increment         (#\esc #\[ Pn 2 #\space #\G)      none      FE)
(SR   scroll-right              (#\esc #\[ Pn #\space #\A)        1         eF)
(SS2  single-shift-2            (#\esc #\N)                       #f        Intro)
(SS3  single-shift-3            (#\esc #\O)                       #f        Intro)

(SSA  start-of-selected-area    (#\esc #\F)                       #f)
(ST   string-terminator         (#\esc #\\)                       #f        Delim)
(STS  set-transmit-state        (#\esc #\S)                       #f)
(SU   scroll-up                 (#\esc #\[ Pn #\S)                1         eF)
(TBC  tab-clear                 (#\esc #\[ Ps #\g)                0         FE)
(TSS  thin-space-specification  (#\esc #\[ Pn #\space #\E)        none      FE)
(VPA  vert-position-absolute    (#\esc #\[ Pn #\d)                1         FE)
(VPR  vert-position-relative    (#\esc #\[ Pn #\e)                1         FE)
(VTS  vertical-tabulation-set   (#\esc #\J)                       #f        FE)

        ))))

(define (term-code name . args)
  (define sep #\;)
  (define (number->revlist n) (reverse! (string->list (number->string n))))
  (let ((data (hash-table-get
               escape-codes
               (if (symbol? name) name (string->symbol name)))))
    (let ((res (list))
          (type #f))
      (for-each
       (lambda (x)
         (cond ((char? x)
                (set! res (cons x res)))
               ((or (eq? x 'Pn) (eq? x 'Ps))
                (set! type x)
                (cond ((pair? args)
                       (set! res (append (number->revlist (car args)) res))
                       (set! args (cdr args)))))
               ((eq? x '*)
                (for-each
                 (lambda (n) (set! res (append (number->revlist n)
                                               (cons sep res))))
                 args)
                (set! args '()))
               ((number? x)
                (do ((i (- x 1) (- i 1)))
                    ((<= i 0) #f)
                  (cond ((pair? args)
                         (set! res (append (number->revlist (car args))
                                           (cons sep res)))
                         (set! args (cdr args))))))
               (else
                (error "unknown terminal code element " x))))
       (car data))
      (list->string (reverse! res)))))

;;; Colors

;; Interface adapted from the Perl Term::ANSIColor module by Zenin
;; (zenin@best.com) and Russ Allbery (rra@stanford.edu).  These are
;; treated separately from other escape codes because you often want to
;; wrap text in between one or more colors and a reset sequence.

;; Color escapes are pretty simple, it's \e[XXm where XX is one or more
;; numeric codes separated by ';', and in particular '0' (\e[0m) is the
;; reset code.

(define color-codes #f)

(cond ((not color-codes)
       (set! color-codes (make-hash-table 'eq?))
       (for-each
        (lambda (c) (hash-table-put! color-codes (car c) (cdr c)))
        '((clear      .  "0")
          (reset      .  "0")
          (bold       .  "1")
          (dark       .  "2")
          (underline  .  "4")
          (underscore .  "4")
          (blink      .  "5")
          (reverse    .  "7")
          (concealed  .  "8")
          (black      . "30")   (on-black   . "40") 
          (red        . "31")   (on-red     . "41") 
          (green      . "32")   (on-green   . "42") 
          (yellow     . "33")   (on-yellow  . "43") 
          (blue       . "34")   (on-blue    . "44") 
          (magenta    . "35")   (on-magenta . "45") 
          (cyan       . "36")   (on-cyan    . "46") 
          (white      . "37")   (on-white   . "47")))))

(define color
  (let ((base (list->string '(#\esc #\[))))
    (lambda args
      (string-append
       base
       (string-join
        (map
         (lambda (x)
           (if (symbol? x)
             (hash-table-get color-codes x)
             (hash-table-get color-codes (string->symbol x))))
         args)
        ";")
       "m"))))

(define reset (color 'reset))

(define (colored . args)
  (string-append (apply color (cdr args)) (car args) reset))

(define (make-color-proc . args)
  (let ((c (apply color args)))
    (lambda (x)
      (string-append c x reset))))

(provide "term/command")

